/*
 * aihandle.c
 */

#include <unistd.h> /* unlink */

#include "aihandle.h"
#include "sgf.h"

/* wait_for_ai
 * checks to see if wait flag is set in either talker
 * returns 1 if so, 0 otherwise
 */
gint wait_for_ai(struct game *g)
{
	if(g->t[BLACK]) {
		if(g->t[BLACK]->waiting)
			return 1;
	}

	if(g->t[WHITE]) {
		if(g->t[WHITE]->waiting)
			return 1;
	}

	return 0;
}

/* cleanup_ai
 * checks both talkers to see if they are alive
 * sends quit, removes them, frees memory
 */
void cleanup_ai(struct game *g)
{

	if(g->t[BLACK]) {
		stop_talk(g->t[BLACK]);
		free_talker(g->t[BLACK]);
		g->t[BLACK] = NULL;
	}
	if(g->t[WHITE]) {
		stop_talk(g->t[WHITE]);
		free_talker(g->t[WHITE]);
		g->t[WHITE] = NULL;
	}

}

/* init_ai
 * initialises ai player for given colour
 * sets player name in game struct to name version
 * returns 1 on failure, 0 if ok
 * Fails if can't get GTP version, set or clear board, or set komi
 * Issues warning but continues if can't set handicap or get name
 */
gint init_ai(struct game *g, gint colour, void(*callback)(void))
{

	gchar *ofilename;
	gchar *nfilename;
	gchar *name;
	gchar *version;
	struct talker *t;

	if(BAD_COLOUR(colour)) {
		g_warning("init_ai: got funny colour\n");
		return 1;
	}

	/* initialise computer player */
	t = new_talker(g->path[colour]->str, g->opts[colour]->str, g->protocol[colour], callback);
	if(!t) {
		g_warning("init_ai: couldn't initialise talker\n");
		return 1;
	}
	if(start_talk(t)) {
		g_warning("init_ai: couldn't start talk\n");
		return 1;
	}

	/* first get protocol version */

	/* NB - this will break a) if there's ever a GTP 3 and b) when we implement GMP,
	 * but that's ok because at that point this will need rewriting anyway, or at least
	 * looking at.
	 */

	if(send_command(t, GT_PVERSION)) {
		g_warning("init_ai: send_command failed - couldn't get GTP version\n");
		return 1;
	}

	if(g_strncasecmp(t->msg, "2", 1) == 0) {
		t->pversion = 2;
	} else if(g_strncasecmp(t->msg, "1", 1) != 0) {
		g_warning("init_ai: unknown GTP version\n");
		return 1;
	}

	if(g->move == 0) {
		/* new game - set boardsize, komi and handicap */
		
		t->idata = g->boardsize;
		
		if(send_command(t, GT_BOARDSIZE)) {
			g_warning("init_ai: send_command failed - couldn't set boardsize\n");
			return 1;
		}
		
		
		/* gtp1 clears board on boardsize, gtp2 does not */
		if(t->pversion == 2) {
			if(send_command(t, GT_CLEAR_BOARD)) {
				g_warning("init_ai: send_command failed: couldn't clear board\n");
				return 1;
			}
		}
		t->fdata = g->komi;
		
		if(send_command(t, GT_KOMI)) {
			g_warning("init_ai: send_command failed - couldn't set komi\n");
			return 1;
		}
		
		/* set handicap if given */
		if(g->handicap > 1) {
			t->idata = g->handicap;
			if(send_command(t, GT_FIXED_HANDICAP)) {
				g_warning("init_ai: send_command failed - couldn't set handicap\n");
				g->handicap = 0;
			}
		}

	} else {
		/* continue game - loadsgf here */
		/* save current sgf file to temporary */
		ofilename = g_strdup(g->col->filename);
		nfilename = g_strconcat(g_get_tmp_dir(), "/goben-tmp.sgf", NULL);
		t->msg = g_strdup(nfilename);
		save_sgf_as(g->col, t->msg);
		g_free(g->col->filename);
		g->col->filename = g_strdup(ofilename);
		/* to get loadsgf to have the current move, give it the next one, hence +1 */
		t->idata = g->move + 1;
		if(send_command(t, GT_LOADSGF)) {
			g_warning("init_ai: send_command failed - couldn't load sgf file\n");
			return 1;
		}
		/* cleanup after ourselves */

		/* delete temporary file here */
		if(unlink(nfilename)) {
			g_warning("init_ai: couldn't unlink %s\n", t->msg);
		}
		g_free(ofilename);
		g_free(t->msg);
		t->msg = NULL;
	}


	/* get name */
	if(send_command(t, GT_NAME)) {
		g_warning("init_ai: send_command failed - couldn't get name\n");
		name = "Name unknown";
	} else {
		name = g_strdup(t->msg);
	}

	if(send_command(t, GT_VERSION)) {
		g_warning("init_ai: send command failed - couldn't get version\n");
		version = "Version unknown";
	} else {
		version = g_strdup(t->msg);
	}

	name = g_strconcat(name, " ", version, NULL);

	g->t[colour] = t;
	g->pname[colour] = g_string_assign(g->pname[colour], name);

	return 0;

}
