/*
 * cfg.c
 */

#include <stdlib.h> /* malloc, free */
#include <stdio.h>  /* fopen */
#include <errno.h>  /* perror */

#include "cfg.h"

/* NB - Must correspond exactly to cfg_items enum */
gchar *cfg_names[NUM_CFG_ITEMS] = {
	"show_coords",
	"coordstyle",
	"hi_contrast",
	"handicap",
	"boardsize",
	"komi",
	"ptypew",
	"ptypeb",
	"pnamew",
	"pnameb",
	"info",
	"protw",
	"protb",
	"pathw",
	"pathb",
	"optsw",
	"optsb",
};

/* get_cfg_filename
 * concatenate user home dir with GOBEN_CFG_FILENAME
 * and return as newly allocated GString
 * caller must free
 */
GString *get_cfg_filename(void)
{

	GString *file;
	/* Construct filename */
	file = g_string_new(g_get_home_dir());
	g_string_append(file, "/");
	g_string_append(file, GOBEN_CFG_FILENAME);

	return file;

}

/* parse_cfg_item
 * attempts to write a single cfg item
 * returns 0 on failure, 1 on success
 * FIXME - needs much error checking to be added
 */
gint parse_cfg_item(struct cfg *c, GString *key, GString *val)
{

	gint i;

	/* If key is valid, assign it to val, and increment count */
	for(i = 0; i < NUM_CFG_ITEMS; i++) {
		if(g_strcasecmp(cfg_names[i], key->str) == 0){
				/* found match */
			switch(i) {
				/* The ints */
			case SHOW_COORDS:
				c->show_coords = atoi(val->str);
				break;
			case COORDSTYLE:
				c->coord_style = atoi(val->str);
				break;
			case HI_CONTRAST:
				c->hi_contrast = atoi(val->str);
				break;
			case HANDICAP:
				c->handicap = atoi(val->str);
				break;
			case BOARDSIZE:
				c->boardsize = atoi(val->str);
				break;
			case PTYPEW:
				c->ptype[WHITE] = atoi(val->str);
				break;
			case PTYPEB:
				c->ptype[BLACK] = atoi(val->str);
				break;
			case PROTW:
				c->protocol[WHITE] = atoi(val->str);
				break;
			case PROTB:
				c->protocol[BLACK] = atoi(val->str);
				break;
				/* The double */
			case KOMI:
				c->komi = g_strtod(val->str, NULL);
				break;
				/* The strings */
			case PNAMEW:
				g_string_assign(c->pname[WHITE], val->str);
				break;
			case PNAMEB:
				g_string_assign(c->pname[BLACK], val->str);
				break;
			case PATHW:
				g_string_assign(c->path[WHITE], val->str);
				break;
			case PATHB:
				g_string_assign(c->path[BLACK], val->str);
				break;
			case OPTSW:
				g_string_assign(c->opts[WHITE], val->str);
				break;
			case OPTSB:
				g_string_assign(c->opts[BLACK], val->str);
				break;
			case INFO:
				g_string_assign(c->info, val->str);
				break;
			};
			
			return 1;
		}
	}

        /* If we get here, we failed to find a key */
	return 0;

}

/* parse_cfg
 * takes cfg struct and file
 * reads file, parses it line by line
 * format is:
 * anything after # is a comment
 * <key>=<value> on a line assigns value to key
 * iff key is recognised (and value appropriate)
 * otherwise does nothing
 * returns 1 if all expected keys found
 * returns 0 otherwise
 */
gint parse_cfg(struct cfg *c, FILE *f)
{

	GString *key;
	GString *val;
	GString *current;
	gint count;
	gint comment;

	gchar ch;
	
	key = g_string_new("");
	val = g_string_new("");

	current = key;

	count = 0;
	comment = 0;

	while(1) {
		ch = getc(f);

		switch(ch) {
		case '#':
			comment = 1;
			break;
		case '\n':
			/* End of line reached */

			/* Attempt to parse the line, record it if we got it right */
			if(parse_cfg_item(c, key, val))
				count++;

			/* Regardless, initialise for new line */
			key = g_string_assign(key, "");
			val = g_string_assign(val, "");
			current = key;
			comment = 0;
			break;
		case EOF:
			/* Free our news */
			g_string_free(key, TRUE);
			g_string_free(val, TRUE);

			/* Make sure we've read everything we were expecting */
			if(count == NUM_CFG_ITEMS)
				return 1;
			else
				return 0;
			break;
		case '=':
			/* when we see an equals sign, switch buf to val */
			if(current != val) {
				current = val;
				break;
			}
			/* otherwise fall through. maybe we want a literal '=' */
		default:
			/* By default, add new character to current buffer */
			if(comment == 0) {
				g_string_append_c(current, ch);
			}
			break;
		};
	}

	/* This should never happen */
	perror("parse_cfg: Reached unreachable section of code");
	return 0;

}

/* cfg2game
 * Writes from cfg struct to game struct
 */
void cfg2game(struct game *g, struct cfg *c)
{

	gint i;

	g->coord_style = c->coord_style;
	g->show_coords = c->show_coords;
	g->hi_contrast = c->hi_contrast;

	for(i = 0; i < 2; i++) {
		g->p[i] = c->ptype[i];
		g_string_assign(g->pname[i], c->pname[i]->str);
		g_string_assign(g->path[i], c->path[i]->str);
		g_string_assign(g->opts[i], c->opts[i]->str);
	}

	g_string_assign(g->info, c->info->str);

	g->handicap = c->handicap;
	g->boardsize = c->boardsize;
	g->komi = c->komi;


}
	


/* reads cfg file
 * fills cfg struct with it
 * copies cfg struct data to game struct
 * 1 for success, 0 for failure
 */
gint load_cfg(struct game *g)
{

	GString *file;
	struct cfg *c;
	FILE *f;

	c = new_cfg();

	file = get_cfg_filename();
	f = fopen(file->str, "r");
	if(!f) {
		perror("load_cfg: fopen failed");
		return 0;
	}

	/* Read the file... */
	if(!parse_cfg(c, f)) {
		perror("load_cfg: Error parsing config file");
		return 0;
	}


	if(fclose(f)) {
		perror("load_cfg: fclose");
		return 0;
	}

	g_string_free(file, TRUE);

	/* Copy data across */
	cfg2game(g, c);

	return 1;


}

/* game2cfg
 * take a game pointer and a cfg pointer
 * copy data from game to cfg
 */
void game2cfg(struct game *g, struct cfg *c)
{

	gint i;

	c->coord_style = g->coord_style;
	c->show_coords = g->show_coords;
	c->hi_contrast = g->hi_contrast;

	for(i = 0; i < 2; i++) {
		c->ptype[i] = g->p[i];
		g_string_assign(c->pname[i], g->pname[i]->str);
		g_string_assign(c->path[i], g->path[i]->str);
		g_string_assign(c->opts[i], g->opts[i]->str);
	}

	g_string_assign(c->info, g->info->str);

	c->handicap = g->handicap;
	c->boardsize = g->boardsize;
	c->komi = g->komi;

}

/* save_cfg
 * takes a game struct
 * fills out an equivalent cfg struct
 * writes the file
 * returns 1 on success, 0 on failure
 */
gint save_cfg(struct game *g)
{

	GString *file;
	struct cfg *c;
	FILE *f;
	gint i;

	/* Copy game data into cfg struct */
	c = new_cfg();
	game2cfg(g, c);

	file = get_cfg_filename();

	f = fopen(file->str, "w");

	if(!f) {
		perror("save_cfg: fopen failed");
		return 0;
	}

	for(i = 0; i < NUM_CFG_ITEMS; i++) {
		/* Write the cfg file here */
		fprintf(f, "%s", cfg_names[i]);
		fprintf(f, "=");

		/* Spot the ugly kluge */
		switch(i) {
			/* All the ints */
		case SHOW_COORDS:
			fprintf(f, "%d", c->show_coords);
			break;
		case COORDSTYLE:
			fprintf(f, "%d", c->coord_style);
			break;
		case HI_CONTRAST:
			fprintf(f, "%d", c->hi_contrast);
			break;
		case HANDICAP:
			fprintf(f, "%d", c->handicap);
			break;
		case BOARDSIZE:
			fprintf(f, "%d", c->boardsize);
			break;
		case PTYPEW:
			fprintf(f, "%d", c->ptype[WHITE]);
			break;
		case PTYPEB:
			fprintf(f, "%d", c->ptype[BLACK]);
			break;
		case PROTW:
			fprintf(f, "%d", c->protocol[WHITE]);
			break;
		case PROTB:
			fprintf(f, "%d", c->protocol[BLACK]);
			break;
			/* The double - one digit precision */
		case KOMI:
			fprintf(f, "%.1f", c->komi);
			break;
			/* The strings */
		case PNAMEW:
			fprintf(f, "%s", c->pname[WHITE]->str);
			break;
		case PNAMEB:
			fprintf(f, "%s", c->pname[BLACK]->str);
			break;
		case PATHW:
			fprintf(f, "%s", c->path[WHITE]->str);
			break;
		case PATHB:
			fprintf(f, "%s", c->path[BLACK]->str);
			break;
		case OPTSW:
			fprintf(f, "%s", c->opts[WHITE]->str);
			break;
		case OPTSB:
			fprintf(f, "%s", c->opts[BLACK]->str);
			break;
		case INFO:
			fprintf(f, "%s", c->info->str);
			break;
 
		};

		fprintf(f, "\n");   

	}

	if(fclose(f)) {
		perror("save_cfg: fclose");
		return 0;
	}

	g_string_free(file, TRUE);

	return 1;

}

/* new_cfg
 * return a newly allocated cfg struct
 */
struct cfg *new_cfg(void)
{
	gint i;
	struct cfg *c;

	c = malloc(sizeof(struct cfg));

	c->info = g_string_new("");

	for(i = 0; i < 2; i++) {
		c->pname[i] = g_string_new("");
		c->path[i] = g_string_new("");
		c->opts[i] = g_string_new("");
	}
	return c;

}

/* free_cfg
 * take a cfg struct
 * free it from memory
 */
void free_cfg(struct cfg *c)
{

	gint i;

	if(c->info)
		g_string_free(c->info, TRUE);

	for(i = 0; i < 2; i++) {
		if(c->pname[i])
			g_string_free(c->pname[i], TRUE);
		if(c->path[i])
			g_string_free(c->path[i], TRUE);
		if(c->opts[i])
			g_string_free(c->opts[i], TRUE);
	}

	free(c);

}
