/*
 * game.c
 */

#include <stdlib.h>

#include "game.h"
#include "sgf.h"
#include "cfg.h"

/* new game
 * initialises a new game struct to default values
 * and returns it
 */
struct game *new_game(void)
{

	struct game *g;

	g = malloc(sizeof(struct game));

	/* basic settings */
	g->rules = GODEF_RULES;
	g->boardsize = GODEF_BOARD;
	g->handicap = GODEF_HANDICAP;
	g->p[BLACK] = GODEF_P1;
	g->p[WHITE] = GODEF_P2;
	g->mode = GODEF_MODE;
	g->komi = GODEF_KOMI;
	g->pname[BLACK] = g_string_new(GODEF_P1NAME);
	g->pname[WHITE] = g_string_new(GODEF_P2NAME);
	g->info = g_string_new(GODEF_INFO);
	g->score = g_string_new("");
	g->comment = g_string_new("");
	g->nodename = g_string_new("");

	/* gui settings */
	g->coord_style = 1;
	g->show_coords = 1;
	g->hi_contrast = 0;

	/* ai settings */
	g->path[BLACK] = g_string_new(GODEF_PATH1);
	g->opts[BLACK] = g_string_new(GODEF_OPTS1);
	g->protocol[BLACK] = GODEF_PROT1;
	g->path[WHITE] = g_string_new(GODEF_PATH2);
	g->opts[WHITE] = g_string_new(GODEF_OPTS2);
	g->protocol[WHITE] = GODEF_PROT2;
	g->t[BLACK] = NULL;
	g->t[WHITE] = NULL;

	/* sgf */
	g->col = NULL;

	/* settings */
	g->need_save = 0;

	/* read cfg file; writes a new one if load failed
	 * (so creating cfg file if it doesn't exist
	 */
	if(!load_cfg(g))
		save_cfg(g);

	init_game(&g, 0);

	return(g);

}

/* free_game
 * frees all memory associated with a game struct
 */
void free_game(struct game *g)
{

	gint i;

	/* free sgf data */
	if(g->col)
		cleanup_sgf(g->col);

	/* free all strings in game struct */
	if(g->info)
		g_string_free(g->info, TRUE);
	if(g->score)
		g_string_free(g->score, TRUE);
	if(g->comment)
		g_string_free(g->comment, TRUE);
	if(g->nodename)
		g_string_free(g->nodename, TRUE);

	/* free pairs of strings etc */
	for(i = 0; i < 2; i++) {
		if(g->pname[i])
			g_string_free(g->pname[i], TRUE);
		if(g->path[i])
			g_string_free(g->path[i], TRUE);
		if(g->opts[i])
			g_string_free(g->opts[i], TRUE);
		if(g->t[i]) 
			free_talker(g->t[i]);
	}

	/* free game struct */
	free(g);

}

/* arrays for handicap setup
 * index into hodd array is number of stones-2,
 * index into odds array is (boardsize/2)-1
 * max of 4 stones on even boards
 * japanese-style arrangement on all odd boards between 9 and 19 inclusive,
 * for up to 9 stones.
 */

/* locations of star points on oddsized boards */
static gint odds[6][3] = {
	{2, 4, 6},
	{2, 5, 8},
	{3, 6, 9},
	{3, 7, 11},
	{3, 8, 13},
	{3, 9, 15},
};

static gint hodd[HANDICAP_MAX-1][HANDICAP_MAX*2] = {
	{3,3, 1,1},
	{3,3, 1,1, 1,3},
	{3,3, 1,1, 1,3, 3,1},
	{3,3, 1,1, 3,1, 1,3, 2,2},
	{3,3, 1,1, 3,1, 1,3, 1,2, 3,2},
	{3,3, 1,1, 3,1, 1,3, 1,2, 3,2, 2,2},
	{3,3, 1,1, 3,1, 1,3, 1,2, 3,2, 2,1, 2,3},
	{3,3, 1,1, 3,1, 1,3, 1,2, 3,2, 2,1, 2,3, 2,2},
};

static gint heven[1][8] = {
	{3,3, 1,1, 1,3, 3,1 },
};

/* set_fixed_handicap
 * put fixed handicap stones on the board
 * demand boardsize >= 9, handicap between 2 and HANDICAP_MAX (9)
 * place maximum of four stones on even boards even if asked for more
 * return 1 on error, 0 otherwise
 */
gint set_fixed_handicap(struct game *g)
{

	gint i;
	gint hx, hy;
	gint hi;     /* index into odds handicap main array */
	gint ho;     /* index into odds handicap location array */
	gint even;   /* how far in the even handicap stones go */

	/* sanity check */
	if((g->handicap < 2) || (g->handicap > HANDICAP_MAX)) {
		g_warning("Handicap setting out of range\n");
		return 1;
	}

	/* boardsize must be >= 7 */
	if(g->boardsize < 7) {
		g_warning("Can't set handicap on board smaller than 7x7\n");
		return 1;
	}

	hi = g->handicap - 2;
	ho = (g->boardsize / 2) - 4;
	even = 2;
	if(g->boardsize > 11)
		even = 3;
 
	for(i = 0; i < (g->handicap * 2); i += 2) {
		/* treat 7x7 board as even, all other odd boards as normal */
		if((g->boardsize != 7) && (g->boardsize & 1)) {
			hx = odds[ho][hodd[hi][i]-1];
			hy = odds[ho][hodd[hi][i+1]-1];
		} else {
			if(i < 8) {
				/* 7x7 - 2,2;4,4; 8x8 - 2,2;5,5; 10x10 - 2,2;7,7; 12x12 - 3,3;8,8
				 * 14x14 - 3,3;10,10; 16x16 - 3,3;12,12; 18x18 - 3,3;14,14
				 */
				if(heven[0][i] == 1) {
					hx = even;
				} else {
					hx = g->boardsize - even - 1;
				}
				if(heven[0][i+1] == 1) {
					hy = even;
				} else {
					hy = g->boardsize - even - 1;
				}

			} else {
				hx = -1;
				hy = -1;
			}
		}
//		g_print("set_fixed_handicap: handicap %d, got %d, %d\n", g->handicap, hx, hy);
		if((hx != -1) && (hy != -1))
			g->board[hx][hy] = GOBAN_B;

	}

	return 0;

}


/* clear_board
 * set all cells to 0
 */
void clear_board(struct game **g)
{

	int i, j;

	for(i = 0; i < (*g)->boardsize; i++) {
		for(j = 0; j < (*g)->boardsize; j++) {
			(*g)->board[i][j] = GOBAN_EMPTY;
		}
	}


}

/* init_game
 * initialise game data
 * clears board, sets up fixed handicap
 * initialises SGF file unless have_sgf is set
 */
void init_game(struct game **g, gint have_sgf)
{

	gint err;

	(*g)->lastx = -1;
	(*g)->lasty = -1;
	(*g)->lastcap = 0;
	(*g)->bcap = 0;
	(*g)->wcap = 0;
	(*g)->move = 0;
	(*g)->pass = 0;
	(*g)->tomove = GOBAN_B;

	(*g)->nodenote = SN_NONE;
	(*g)->movenote = SM_NONE;
	(*g)->nodeval = 0.0;

	clear_board(g);

	if((*g)->handicap > 1) {
		err = set_fixed_handicap(*g);
		if(err) {
			g_warning("init_game: problem setting handicap\n");
		}
		(*g)->tomove = GOBAN_W;
	}

	/* Initialise SGF stuff */
	if(have_sgf == 0) {
		(*g)->have_result = 0;
		(*g)->num_moves = 0;
		if((*g)->col) {
			cleanup_sgf((*g)->col);
		}
		
		(*g)->col = init_sgf();
		set_sgf_header(*g);
	}
	
}

/* toggle_tomove
 * toggle move flag
 */
void toggle_tomove(struct game *g)
{

	if(g->tomove == GOBAN_W) {
		g->tomove = GOBAN_B;
	} else {
		g->tomove = GOBAN_W;
	}

}

/* to_play
 * returns WHITE if given GOBEN_W
 * returns BLACK if given GOBEN_B
 */
gint col_to_play(gint stone)
{
	
	if(stone == GOBAN_W)
		return WHITE;

	if(stone != GOBAN_B) {
		g_warning("col_to_play - given weird input %d\n", stone);
	}

	return BLACK;

}

gint stone_to_playnomore(gint move)
{

	if(col_to_play(move) == BLACK) {
		return GOBAN_B;
	}
		
	return GOBAN_W;

}

/* clear_count
 * clear up after counting liberties
 * if dead is GOBAN_W or GOBAN_B, assume that any seen stones
 * of that colour are to be marked dead
 */
gint clear_count(struct game **g, gint dead)
{
	gint i, j;
	gint w, b;
	gint cap = 0;

	w = GOBAN_W;
	b = GOBAN_B;

	if(dead == GOBAN_W) {
		w = GOBAN_DW;
	} else if(dead == GOBAN_B) {
		b = GOBAN_DB;
	}

	for(i = 0; i < (*g)->boardsize; i++) {
		for(j = 0; j < (*g)->boardsize; j++) {
			switch((*g)->board[i][j]) {
			case GOBAN_LIB:
				(*g)->board[i][j] = GOBAN_EMPTY;
				break;
			case GOBAN_SW:
				(*g)->board[i][j] = w;
				if(w == GOBAN_DW)
					cap++;
				break;
			case GOBAN_SB:
				(*g)->board[i][j] = b;
				if(b == GOBAN_DB)
					cap++;
				break;
			case GOBAN_SDW:
				(*g)->board[i][j] = GOBAN_DW;
				break;
			case GOBAN_SDB:
				(*g)->board[i][j] = GOBAN_DB;
				break;
			default:
				break;
			}
		}
	}

	return cap;

}

/* count_liberties
 * returns the number of liberties for the stone/group at x,y
 * recursive function, leaves a trail of seen cells that must
 * be cleaned up elsewhere
 */
gint count_liberties(struct game *g, gint x, gint y)
{

	gint lib = 0;
	gint this = g->board[x][y];
	gint dead_opp, seen_dead_opp;

	/* sanity check */
	if((this != GOBAN_W) &&
	   (this != GOBAN_B)) {
		g_warning("Non live cell passed to count liberties\n");
		return lib;
	}

//	g_print("count_liberties at %d, %d - boardsize %d\n", x, y, g->boardsize);

	/* set this cell seen */
	if(this == GOBAN_W) {
		g->board[x][y] = GOBAN_SW;
		dead_opp = GOBAN_DB;
		seen_dead_opp = GOBAN_SDB;
	} else {
		g->board[x][y] = GOBAN_SB;
		dead_opp = GOBAN_DW;
		seen_dead_opp = GOBAN_SDB;
	}

	if(x > 0) {

//		g_print("checking %d, %d\n", x-1, y);

		if (g->board[x-1][y] == GOBAN_EMPTY) {
			lib++;
			g->board[x-1][y] = GOBAN_LIB;
		} else if (g->board[x-1][y] == dead_opp) {
			lib++;
			g->board[x-1][y] = seen_dead_opp;
		} else if (g->board[x-1][y] == this) {
			lib += count_liberties(g, x-1, y);
		}
	}

	if(x < g->boardsize - 1) {

//		g_print("checking %d, %d\n", x+1, y);

		if (g->board[x+1][y] == GOBAN_EMPTY) {
			lib++;
			g->board[x+1][y] = GOBAN_LIB;
		} else if (g->board[x+1][y] == dead_opp) {
			lib++;
			g->board[x+1][y] = seen_dead_opp;
		} else if (g->board[x+1][y] == this) {
			lib += count_liberties(g, x+1, y);
		}
	}

	if(y > 0) {

//		g_print("checking %d, %d\n", x, y-1);

		if (g->board[x][y-1] == GOBAN_EMPTY) {
			lib++;
			g->board[x][y-1] = GOBAN_LIB;
		} else if (g->board[x][y-1] == dead_opp) {
			lib++;
			g->board[x][y-1] = seen_dead_opp;
		} else if (g->board[x][y-1] == this) {
			lib += count_liberties(g, x, y-1);
		}
	}

	if(y < g->boardsize - 1) {

//		g_print("checking %d, %d\n", x, y+1);

		if (g->board[x][y+1] == GOBAN_EMPTY) {
			lib++;
			g->board[x][y+1] = GOBAN_LIB;
		} else if (g->board[x][y+1] == dead_opp) {
			lib++;
			g->board[x][y+1] = seen_dead_opp;
		} else if (g->board[x][y+1] == this) {
			lib += count_liberties(g, x, y+1);
		}
	}

//	g_print("count_liberties: returning %d\n", lib);

	return lib;

}

/* do_set
 * arbitrarily sets x,y to given colour (or empty)
 * returns 0 if ok, 1 if out of range
 */
gint do_set(struct game **g, gint x, gint y, gint colour)
{
	/* check x and y are in range */
	if((x < 0) || (x >= (*g)->boardsize) ||
	   (y < 0) || (y >= (*g)->boardsize)) {
		g_warning("x,y val for move %d out of range (%d, %d)\n", (*g)->move, x, y);
		return GOBAN_ERR_RANGE;
	}

	(*g)->board[x][y] = colour;

	return 0;

}

/* do_move
 * returns 0 if ok
 * returns a GOBAN_ERR_* otherwise
 * implements captures, disallows suicide, disallows ko
 * increments move and capture counters
 * sets last move and last capture
 */
gint do_move(struct game **g, gint x, gint y)
{

	gint stone;
	gint opp;
	gint captured = 0;
	gint ko = 0;

	/* allow any -1 to be pass */
	if((x == -1) || (y == -1)) {
		(*g)->move++;
		toggle_tomove(*g);
		return 0;
	}

	/* check x and y are in range */
	if((x < 0) || (x >= (*g)->boardsize) ||
	   (y < 0) || (y >= (*g)->boardsize)) {
		g_warning("x,y val for move %d out of range (%d, %d)\n", (*g)->move, x, y);
		return GOBAN_ERR_RANGE;
	}
	 
	/* check x,y on board is unoccupied */
	if((*g)->board[x][y] != GOBAN_EMPTY) {
		g_warning("x,y val for move %d - already occupied(%d) (%d, %d)\n",
			  (*g)->move, (*g)->board[x][y], x, y);
		return GOBAN_ERR_OCC;
	}

	/* find who's turn it is to play */
	stone = (*g)->tomove;
	/* put the stone down */
	(*g)->board[x][y] = stone;
	/* find the stone of the opposite colour */
	toggle_tomove(*g);
	opp = (*g)->tomove;

	/* check for captures */
	if(x > 0) {
		if((*g)->board[x-1][y] == opp) {
			if(count_liberties(*g, x-1, y) == 0) {
				/* check for ko */
				if(((*g)->lastx == x-1) &&
				   ((*g)->lasty == y) &&
				   ((*g)->lastcap == 1)) {
					clear_count(g, GOBAN_EMPTY);
					ko++;
				} else {
					captured += clear_count(g, opp);
				}
			} else {
				clear_count(g, GOBAN_EMPTY);
			}
		}
	}
	if(x < (*g)->boardsize) {
		if((*g)->board[x+1][y] == opp) {
			if(count_liberties(*g, x+1, y) == 0) {
				if(((*g)->lastx == x+1) &&
				   ((*g)->lasty == y) &&
				   ((*g)->lastcap == 1)) {
					clear_count(g, GOBAN_EMPTY);
					ko++;
				} else {
					captured += clear_count(g, opp);
				}
			} else {
				clear_count(g, GOBAN_EMPTY);
			}
		}
	}
	if(y > 0) {
		if((*g)->board[x][y-1] == opp) {
			if(count_liberties(*g, x, y-1) == 0) {
				if(((*g)->lastx == x) &&
				   ((*g)->lasty == y-1) &&
				   ((*g)->lastcap == 1)) {
					clear_count(g, GOBAN_EMPTY);
					ko++;
				} else {
					captured += clear_count(g, opp);
				}
			} else {
				clear_count(g, GOBAN_EMPTY);
			}
		}
	}
	if(y < (*g)->boardsize) {
		if((*g)->board[x][y+1] == opp) {
			if(count_liberties(*g, x, y+1) == 0) {
				if(((*g)->lastx == x) &&
				   ((*g)->lasty == y+1) &&
				   ((*g)->lastcap == 1)) {
					clear_count(g, GOBAN_EMPTY);
					ko++;
				} else {
					captured += clear_count(g, opp);
				}
			} else {
				clear_count(g, GOBAN_EMPTY);
			}
		}
	}

	/* now we can check for suicide */
	if(count_liberties(*g, x, y) == 0) {
		/* still no liberties? can't move here */	
		toggle_tomove(*g);
		(*g)->board[x][y] = GOBAN_EMPTY;
		clear_count(g, GOBAN_EMPTY);
		if(ko) {
			return GOBAN_ERR_KO;
		} else {
			return GOBAN_ERR_SUI;
		}
	}
	clear_count(g, GOBAN_EMPTY);

	/* increment counters */
	(*g)->move++;
	if(stone == GOBAN_W) {
		(*g)->wcap += captured;
	} else {
		(*g)->bcap += captured;
	}

	/* set info for next go's ko check */
	(*g)->lastx = x;
	(*g)->lasty = y;
	(*g)->lastcap = captured;

	return 0;
}

/* remove_captured
 * scan board for captured stones, set them empty
 */
void remove_captured(struct game **g)
{

	gint i, j;

	for(i = 0; i < (*g)->boardsize; i++) {
		for(j = 0; j < (*g)->boardsize; j++) {
			if(((*g)->board[i][j] == GOBAN_DW) ||
			   ((*g)->board[i][j] == GOBAN_DB)) {
				(*g)->board[i][j] = GOBAN_EMPTY;
			}
		}
	}

} 

