/*
 * game.h
 */

#ifndef __GAME_H__
#define __GAME_H__

#include <glib.h>

#include "config.h"
#include "talk.h"
#include "sgftree.h"

/* default settings you might want to edit */
#define GNUGO_PATH "/usr/local/bin/gnugo"
#define GNUGO_OPTIONS "--mode gtp --quiet --level 1"

/* limits */
#define BOARD_MIN 5
#define BOARD_MAX 19
/* nb - not sure about these */
#define KOMI_MIN -120.0
#define KOMI_MAX 120.0 /* ?? *really* dunno */
#define HANDICAP_MIN 0
#define HANDICAP_MAX 9

/* enums */
enum rules { GOBAN_JAPANESE, GOBAN_CHINESE };
enum gomodes { GOBAN_PLAYH, GOBAN_PLAYC, GOBAN_PLAYW, GOBAN_SCORE, GOBAN_OVER, GOBAN_EDIT, GOBAN_REPLAY, GOBAN_ERROR };
enum ptypes { GOBAN_HUMAN, GOBAN_AI };
enum cells { GOBAN_EMPTY, GOBAN_W,  GOBAN_B,  GOBAN_DW, GOBAN_DB,
	     GOBAN_LIB,   GOBAN_SW, GOBAN_SB, GOBAN_SDW, GOBAN_SDB };
enum errors { GOBAN_ERR_NONE, GOBAN_ERR_RANGE, GOBAN_ERR_OCC, GOBAN_ERR_SUI, GOBAN_ERR_KO };

enum nodenotes { SN_NONE, SN_DM1, SN_DM2, SN_GB1, SN_GB2, SN_GW1, SN_GW2, SN_HO1, SN_HO2, SN_UC1, SN_UC2 };
enum movenotes { SM_NONE, SM_BM1, SM_BM2, SM_DO, SM_IT, SM_TE1, SM_TE2 };

/* default settings */
#define GODEF_RULES GOBAN_JAPANESE
#define GODEF_BOARD 19
#define GODEF_HANDICAP 0
#define GODEF_P1 GOBAN_HUMAN
#define GODEF_P2 GOBAN_AI
#define GODEF_MODE GOBAN_EDIT
#define GODEF_KOMI 5.5
#define GODEF_P1NAME "Black"
#define GODEF_P2NAME "White"
#define GODEF_INFO "Goben v" VERSION
#define GODEF_PATH1 GNUGO_PATH
#define GODEF_PATH2 GNUGO_PATH
#define GODEF_OPTS1 GNUGO_OPTIONS
#define GODEF_OPTS2 GNUGO_OPTIONS
#define GODEF_PROT1 GT_GTP
#define GODEF_PROT2 GT_GTP

#define BLACK 0
#define WHITE 1

#define BAD_COLOUR(x) (((x) != BLACK) && ((x) != WHITE))

/* game struct - probably missing stuff */
struct game {
	gint rules;          /* what ruleset are we playing */
	gint boardsize;      /* clamp between 5-19, like gnugo */
	gint handicap;       /* number of handicap stones given */
	gint move;           /* number of move */
	gint tomove;         /* who is about to move */
	gint mode;           /* play, edit, replay, error */
	gint pass;           /* number of passes in a row. game ends when this hits two */
	gint lastx;          /* x coord of last move */
	gint lasty;          /* y coord of last move */
	gint lastcap;        /* how many stones last move captured */
	gint bcap;           /* how many stones has black captured */
	gint wcap;           /* how many stones has white captured */
	gint have_result;    /* for sgf files, whether we have result */
	gint num_moves;      /* for sgf files, number of moves found */
	gdouble komi;        /* komi value*/
	gint board[19][19];  /* bet there's a better way to do this */
	gint p[2];           /* player types 0 - human, 1 - ai */
	gint protocol[2];    /* for ai players, protocol to use */

	/* GUI settings */
	gint coord_style;
	gint show_coords;
	gint hi_contrast;

	/* SGF move/node annotation data */
	gint movenote;
	gint nodenote;
	gdouble nodeval;
	GString *nodename;   /* a string for named SGF nodes */
	GString *comment;    /* a string for comments */

	GString *info;       /* an arbitrary string, defaults to Goben name and version */
	GString *score;      /* a string for the score */
	GString *pname[2];   /* player names */
	GString *path[2];    /* paths to ai players */
	GString *opts[2];    /* options for ai players */
	struct talker *t[2]; /* connections to ai players */
	struct sgfcoll *col; /* SGF data structure */

	gint need_save;      /* Remember if settings saved or not */
};

/* game methods */
struct game *new_game(void);
void free_game(struct game *g);

void init_game(struct game **g, gint have_sgf);
void clear_board(struct game **g);     /* set all cells to 0 */
void toggle_tomove(struct game *g);
gint col_to_play(gint move); /* find whose turn it is to play */
gint do_move(struct game **g, gint x, gint y);
gint do_set(struct game **g, gint x, gint y, gint colour);

void remove_captured(struct game **g);  /* remove captured stones */


#endif /* __GAME_H__ */
