/*
 * gtp.c
 */

#include <unistd.h> /* read, write */
#include <stdio.h>  /* sscanf */
#include <ctype.h>  /* tolower */
#include <stdlib.h> /* malloc */
#include <errno.h>  /* errno */

#include "gtp.h"
#include "game.h"

#define GTP_MAXBUF 1000

gchar *commands[GT_NUM_COMMANDS] = {
	"quit",
	"final_score",
	"name",
	"version",
	"boardsize",
	"komi",
	"fixed_handicap",
	"protocol_version",
	"clear_board",
	"loadsgf"
};

gchar *colours[2] = {
	"black",
	"white",
};

/* _gtp_send
 * actually perform a write on a file descriptor
 */
gint _gtp_send(int fd_out, gchar *msg)
{
	gint n;
	gint err = 0;

	n = write(fd_out, msg, strlen(msg));

	if(n == -1) {
		g_warning("_gtp_send: write failed\n");
		err = n;
	}

	return err;

}

/* _gtp_listen
 * perform a read on a nonblocking file descriptor
 * while waiting, call callback
 */
gchar *_gtp_listen(int fd_in, void(*callback)(void), gint *wait)
{
	gint n;
	static gchar buf[GTP_MAXBUF];

	*wait = 1;

	while(*wait) {

		n = read(fd_in, buf, GTP_MAXBUF);
		
		if(n == -1) {
			/* read pipe now nonblocking, maybe we just need to wait */
			if(errno == EAGAIN) {
				/* prod gtk here */
				callback();
				continue;
			}
			g_warning("_gtp_listen: read failed\n");
			*wait = 0;
			return NULL;
		} else {
			*wait = 0;
		}
		
	}

	/* this shouldn't happen, should it? */
	if(n == GTP_MAXBUF) {
		g_warning("_gtp_listen: read max no of bytes\n");
		n--; /* we're about to access buf[n] */
	}

	/* make sure string is properly terminated whatever */
	buf[n] = '\0';

	/* is this ok? */
	return buf;

}

/* _gtp_parse
 * parse a response into string form
 * put the string in *pmsg
 * return 0 if ok, 1 for error message
 * return -1 if we couldn't parse it
 */
gint _gtp_parse(gchar *msg, gchar **pmsg)
{

	gint err = 0;
	gchar *tmp;

	if(!msg)
		g_warning("_gtp_parse: msg was empty string\n");

	if(msg[0] == '?') {
		err = 1;
	} else if (msg[0] != '=') {
		g_warning("_gtp_parse: got weird message '%s'\n", msg);
		return -1;
	}
	
	//	g_print("_gtp_parse: Got '%s'\n", msg);

	tmp = msg+1;
	g_strstrip(tmp);

	//	g_print("_gtp_parse: Now '%s'\n", tmp);

	*pmsg = g_strdup(tmp);

	return err;

}
/* _gtp_chat
 * handle a command/response sequence
 */
gint _gtp_chat(struct talker *t, gchar *cmd)
{

	gint err;
	gchar *rsp;

	if(t->waiting) {
		g_warning("_gtp_chat: already waiting for command\n");
		return -1;
	}

	if(t->state != GT_TALKING) {
		g_warning("_gtp_chat: not connected\n");
		return -1;
	}

	/* send command */
	if(_gtp_send(t->p->fd_out, cmd)) {
		g_warning("_gtp_chat: write failed\n");
		return -1;
	}

	/* get response */
	rsp = _gtp_listen(t->p->fd_in, t->callback, &(t->waiting));

	err = _gtp_parse(rsp, &t->msg);
	
	if(err == 1) {
		g_warning("_gtp_chat: received gtp error %s\n", t->msg);
	}

	return err;

}

/* _gtp_movetoi
 * pass it a string and two pointers to int
 * string read as move and converted accordingly
 * -1,-1 denotes a pass
 * -2,-2 denotes resignation
 */
gint _gtp_movetoi(gchar *move, gint *i, gint *j)
{

	char column;
	int row;
	int n;
	
	//	g_print("_gtp_movetoi: got %s\n", move);

	/* handle passes */
	if(g_strncasecmp(move, "PASS", 4) == 0) {
		*i = -1;
		*j = -1;
		return 0;
	}
	/* handle resignation
	 * NB - I've never seen this happen, it may be wrong... ;)
	 */
	if(g_strncasecmp(move, "RESIGN", 6) == 0) {
		*i = -2;
		*j = -2;
		return 0;
	}

	if (sscanf(move, " %c%d%n", &column, &row, &n) != 2) {
		g_warning("_gtp_movetoi couldn't read %s\n", move);
		return -1;
	}
	
	/* gtp omits 'i', unlike, say, gtkgoboard... */
	if (tolower((int) column) == 'i') {
		g_warning("_gtp_movetoi: got given column 'i'\n");
		return -1;
	}
	*i = tolower((int) column) - 'a' + 1;
	if (tolower((int) column) > 'i')
		--*i;

	*j = row;

	//	g_print("_gtp_movetoi: converted to %d, %d", *i, *j);

	return 0;

}

/* _gtp_itomove
 * send it two ints
 * get back a converted string
 * translate -1,-1 to 'pass'
 */
gchar *_gtp_itomove(gint i, gint j)
{

	gchar *move;
	move = malloc(sizeof(char) * 5);

	if((i == -1) && (j == -1)) {
		snprintf(move, 5, "pass");
	} else {
		snprintf(move, 5, "%c%d", 'A' + i - 1 + (i >= 9), j);
	}
	//	g_print("_gtp_itomove: got %d, %d\n", i, j);
	//      g_print("_gtp_itomove: converted to %s\n", move);

	return move;

}

/* gtp_send_command
 * send appropriate command
 * response appears in t->msg
 */
gint gtp_send_command(struct talker *t, gint command)
{

	gint err;
	gint argslen;
	gchar *cmd;
	gchar *args;

	if(t->state != GT_TALKING) {
		g_warning("gtp_send_command: not connected\n");
		return -1;
	}

	/* create command */
	if(command == GT_LOADSGF) {

		/* move number in idata, filename in msg */
		/* arbitrary but reasonable upper limit on move number */
		if((t->idata <= 0) || (t->idata > 9999)) {
			g_warning("gtp_send_command: bad move number\n");
			return -1;
		}
		if(!t->msg) {
			g_warning("gtp_send_command: no filename provided to loadsgf\n");
			return -1;
		}
		argslen = strlen(t->msg) + 7;
		args = malloc(sizeof(char) * argslen);
		snprintf(args, argslen, " %s %d", t->msg, t->idata);
		cmd = g_strconcat(commands[command], args, "\n", NULL);
		g_free(args);

	} else if(command == GT_BOARDSIZE) {
		if((t->idata < BOARD_MIN) || (t->idata > BOARD_MAX)) {
			g_warning("gtp_send_command: bad boardsize\n");
			return -1;
		}
		args = malloc(sizeof(char) * 4);
		snprintf(args, 4, " %d", t->idata);
		cmd = g_strconcat(commands[command], args, "\n", NULL);
		g_free(args);
	} else if (command == GT_FIXED_HANDICAP) {
		if((t->idata < HANDICAP_MIN) || (t->idata > HANDICAP_MAX)) {
			g_warning("gtp_send_command: bad handicap\n");
			return -1;
		}
		args = malloc(sizeof(char) * 3);
		snprintf(args, 3, " %d", t->idata);
		cmd = g_strconcat(commands[command], args, "\n", NULL);
		g_free(args);
	} else if (command == GT_KOMI) {
		if((t->fdata < KOMI_MIN) || (t->fdata > KOMI_MAX)) {
			g_warning("gtp_send_command: bad komi\n");
			return -1;
		}
		args = malloc(sizeof(char) * 6);
		snprintf(args, 6, " %.2f", t->fdata);
		cmd = g_strconcat(commands[command], args, "\n", NULL);
	} else { 
		cmd = g_strconcat(commands[command], "\n", NULL);
	}

//	g_print("gtp_send_command: sending %s\n", cmd);

	/* send it */
	err = _gtp_chat(t, cmd);

	/* forget it */
	g_free(cmd);

//	g_print("gtp_send_command: got %s\n", t->msg);

	return err;

}

/* gtp_send_move
 * construct and send a move
 */
gint gtp_send_move(struct talker *t, gint i, gint j, gint colour)
{

	gint err;
	gchar *cmd;
	gchar *move;

	/* are you still talking to me? */
	if(t->state != GT_TALKING) {
		g_warning("gtp_send_move: not connected\n");
		return -1;
	}

	/* have we got a funny colour? */
	if((colour < 0) ||
	   (colour > 1)) {
		g_warning("gtp_send_move: passed peculiar colour value, bailing\n");
		return -1;
	}

	move = _gtp_itomove(i, j);

	/* handle gtp1 and gtp2, assume gtp>2 will behave like gtp2 */
	if(t->pversion == 1) {
		cmd = g_strconcat(colours[colour], " ", move, "\n", NULL);
	} else {
		cmd = g_strconcat("play ", colours[colour], " ", move, "\n", NULL);
	}

	//	g_print("gtp_send_move: sending %s\n", cmd);

	err = _gtp_chat(t, cmd);
	if(err) {
		g_warning("gtp_send_move: got error with cmd %s\n", cmd);
	}

	g_free(cmd);
	g_free(move);

	return err;
}

gint gtp_get_move(struct talker *t, gint *i, gint *j, gint colour)
{

	gint err;
	gchar *cmd;

	if(t->state != GT_TALKING) {
		g_warning("gtp_get_move: not connected\n");
		return -1;
	}

	if((colour < 0) || (colour > 1)) {
		g_warning("gtp_get_move: passed peculiar colour value, bailing\n");
		return -1;
	}

	/* handle gtp1 and gtp2, assume gtp>2 will be like gtp2 */
	if(t->pversion == 1) {
		cmd = g_strconcat("genmove_", colours[colour], "\n", NULL);
	} else {
		cmd = g_strconcat("genmove ", colours[colour], "\n", NULL);
	}

	err = _gtp_chat(t, cmd);

	if(err == 0) {
		/* convert coordinate value in t->msg */
		err = _gtp_movetoi(t->msg, i, j);
	}

	return err;
}
