/*
 * piper.c
 */

#include <stdlib.h> /* malloc, free */
#include <unistd.h> /* close */
#include <fcntl.h>  /* uh, fcntl */

#include "goben.h"
#include "piper.h"

/* Is the following either sensible or necessary? */
#define IN 0
#define OUT 1
#define STDIN IN
#define STDOUT OUT

/* new_piper
 * create and return a new piper struct
 */
struct piper *new_piper(gchar *app, gchar *options)
{

	struct piper *p;

	p = malloc(sizeof(struct piper));

	p->err = 0;
	p->connected = 0;
	p->fd_in = -1;
	p->fd_out = -1;

	/* FIXME - should check that app is sane and return error if not */

	p->app = g_strdup(app);
	p->options = g_strdup(options);

	return p;

}

/* free_piper
 * calls close if we appear to be connected
 * frees memory associated with a piper struct
 */
void free_piper(struct piper *p)
{

	/* this should probably generate a warning */
	/* FIXME - close_pipe may fail. What should we do? */
	if(p->connected)
		close_pipe(p);

	if(p->app)
		g_free(p->app);

	if(p->options)
		g_free(p->options);

	free(p);

}

/* close_pipe
 * if we appear to be connected, close the fds
 * and set p->connected to 0
 * returns 0 if ok
 * returns -1 if one of the fds wouldn't close
 * returns -2 if neither would close
 */
gint close_pipe (struct piper *p)
{

	gint err = 0;
	gint err1, err2;

	if(p->connected) {
		err1 = close(p->fd_in);
		err2 = close(p->fd_out);
		err = err1 + err2;
		if(!err)
			p->connected = 0;
	}

	return err;

}

/* connect_pipe
 * probably misnamed
 * open two pipes
 * fork off a child
 * child dups stdin and stout to the pipes, then execs app
 * we return, hopefully, 0, or -1 if an error occured
 *
 * pipe(a2b[2]) - a2b[0] is for reading, a2b[1] is for writing
 *
 */
gint connect_pipe(struct piper *p)
{
	int fd;
	pid_t pid;
	gint err = 0;
	gint gui2prog[2];
	gint prog2gui[2];
	gchar **opts;
	gchar *optstr;
	glong flags;

	/* outbound pipe */
	err = pipe(gui2prog);
	if(err == -1)
		return err;

	/* inbound pipe */
	err = pipe(prog2gui);
	if(err == -1)
		return err;

	/* fork and exec */
	pid = fork();

	if(pid == -1) {
		/* Could not fork */
		g_warning("connect_pipe: fork() failed\n");
		return -1;
	} else if(pid == 0) {
		/* child */

		/* dup stdin and stdout, close other fds */
		/* NB - g_warning *does* send to stderr, right? */
		for (fd = 0;  fd < 100;  ++fd)  {
			if ((fd != 2) &&  /* Don't close stderr! */
			    (fd != gui2prog[0]) && (fd != prog2gui[1]))
				close(fd);
		}


		err = dup(gui2prog[0]);
		if(err == -1) {
			g_warning("connect_pipe: child couldn't dup stdin, bailing\n");
			_exit(1);
		}
		err = close(gui2prog[0]);
		if(err) {
			/* What should we do about such an error? */
			g_warning("connect_pipe: child couldn't close gui2prog[0]\n");
		}



		err = dup(prog2gui[1]);
		if(err == -1) {
			g_warning("connect_pipe: child couldn't dup stdout, bailing\n");
			_exit(1);
		}
		err = close(prog2gui[1]);
		if(err) {
			/* Ditto previous comment */
			g_warning("connect_pipe: child couldn't close prog2gui[1]\n");
		}


		

		optstr = g_strconcat(g_basename(p->app), " ", p->options, NULL);
		opts = g_strsplit(optstr, " ", 0);

		execv(p->app, opts);

		/* If execv returns, we have a problem */
		g_warning("connect (in child): execv() failed\n");
		_exit(1);

	} else {
		/* parent */

		//		g_print("connect_pipe: (in parent) - fork seems ok here\n");

	}

	/* set up struct, close unnecessary fds */
	p->fd_in = prog2gui[IN];
	err = close (prog2gui[OUT]);
	if(err)
		return err;

	/* set prog2gui[1] nonblocking so we don't freeze when reading it */
	flags = fcntl(prog2gui[IN], F_GETFL);
	err = fcntl(prog2gui[IN], F_SETFL, flags | O_NONBLOCK);
	if(err) {
		g_warning("connect_pipe: could not set prog2gui[IN] non-blocking\n");
	}


	p->fd_out = gui2prog[OUT];
	err = close(gui2prog[IN]);
	if(err)
		return err;

	p->connected = 1;

	return err;

}


