/*
 * talk.c
 */

#include <stdlib.h> /* malloc, free */
#include "talk.h"
#include "gtp.h"

/* new_talker
 * initialise the talk struct
 * sets state to READY
 * only supports GTP1 mode for now
 */
struct talker *new_talker(gchar *app, gchar *options, gint mode, void(*callback)(void))
{

        struct talker *t;

	if(mode != GT_GTP) {
		g_warning("new_talker: non GTP mode requested, bailing\n");
		return NULL;
	}

	t = malloc(sizeof(struct talker));

	t->p = new_piper(app, options);
	t->mode = mode;
	t->state = GT_READY;
	t->msg = NULL;
	t->pversion = 1; /* default setting */
	t->callback = callback; /* callback so read can be nonblocking */
	t->waiting = 0;

	return t;

}

/* free_talker
 * clean up after a talker
 * emits warning if was in GT_TALKING state
 */
void free_talker(struct talker *t)
{

	if(t) {
		if(t->state == GT_TALKING) {
			send_command(t, GT_QUIT);
		}
		
		if(t->msg)
			free(t->msg);
		
		free_piper(t->p);
		free(t);
	}

}

/* start_talk
 * fail unless state is GT_READY
 * initialise appropriate connection
 * return 0 if ok, -1 on error
 */
gint start_talk(struct talker *t)
{

	gint err = 0;

	if(t->state != GT_READY) {
		g_warning("start_talk: talker was not ready\n");
		return -1;
	}

	switch(t->mode) {
	case GT_GTP:
		/* initialise gtp */
		if(connect_pipe(t->p)) {
			g_warning("start_talk: connect_pipe failed\n");
			t->state = GT_ERROR;
			err = -1;
		}
		t->state = GT_TALKING;
		
		/* GTP initialisation handled in aihandle.c */

		break;
	default:
		g_warning("start_talk: unknown mode\n");
		t->state = GT_ERROR;
		err = -1;
		break;
	}

	return err;

}

/* stop_talk
 * send quit message and disconnect
 */
gint stop_talk(struct talker *t)
{
	gint err = 0;

	if(t->state == GT_READY) {
		g_warning("stop_talk: talker was in ready mode");
		return -1;
	}

	switch(t->mode) {
	case GT_GTP:

		/* Send quit message */
		send_command(t, GT_QUIT);

		err = close_pipe(t->p);
		
		if(err) {
			g_warning("stop_talk: error closing pipes\n");
			t->state = GT_ERROR;
		} else {
			t->state = GT_READY;
		}
		
		break;
	default:
		g_warning("stop_talk: unknown mode\n");
		t->state = GT_ERROR;
		err = -1;
	}

	return err;
}

/* send_move
 * send a move for the given colour
 */
gint send_move(struct talker *t, gint x, gint y, gint colour)
{
	gint err = 0;
	
	if(t->state != GT_TALKING) {
		g_warning("send_move: talker not connected\n");
		return -1;
	}
	
	switch(t->mode) {
	case GT_GTP:
		
//		g_print("send_move: calling gtp_send_move at %d,%d\n", x,y);

		if(gtp_send_move(t, x, y, colour)) {
			g_warning("send_move: gtp_send_move returned nonzero\n");
		}
		
		break;
	default:
		g_warning("send_move: unknown mode\n");
		t->state = GT_ERROR;
		err = -1;
	}
	
	return err;
	
}

/* get_move
 * get a move for the given colour
 */
gint get_move(struct talker *t, gint *x, gint *y, gint colour)
{
  
	gint err = 0;
	
	if(t->state != GT_TALKING) {
		g_warning("get_move: talker not connected\n");
		return -1;
	}
	
	switch(t->mode) {
	case GT_GTP:
		
		if(gtp_get_move(t, x, y, colour)) {
			g_warning("get_move: gtp_get_move returned nonzero\n");
		}
		
		break;
	default:
		g_warning("get_move: unknown mode\n");
		t->state = GT_ERROR;
		err = -1;
	}
	
	return err;
	
}

/* send_command
 * send one of the known generic commands
 */
gint send_command(struct talker *t, gint command)
{

	gint err = 0;

	if(t->state != GT_TALKING) {
		g_warning("send_command: talker not connected\n");
		return -1;
	}

	if((command < 0) ||
	   (command >= GT_NUM_COMMANDS)) {
		g_warning("send_command: unknown command\n");
		/* should this not set t->state to error? */
		return -1;
	}

	switch(t->mode) {
	case GT_GTP:

		if(gtp_send_command(t, command)) {
			g_warning("send_command: command failed\n");
			err = -1;
		}

		break;
	default:
		g_warning("send_command: unknown mode\n");
		t->state = GT_ERROR;
		err = -1;
	}

	return err;


}
